@@ -5,7 +5,7 @@
 
 package Bio::Kmer;
 require 5.10.0;
-our $VERSION=0.23;
+our $VERSION=0.24;
 
 use strict;
 use warnings;
@@ -16,6 +16,8 @@ use File::Temp qw/tempdir tempfile/;
 use File::Path qw/remove_tree/;
 use Data::Dumper qw/Dumper/;
 use IO::Uncompress::Gunzip;
+use File::Which qw/which/;
+use Carp qw/croak carp confess/;
 
 use threads;
 use threads::shared;
@@ -151,10 +153,11 @@ sub new{
     my $tempfile="$$settings{tempdir}/bioperl_input.fastq";
     my $out=Bio::SeqIO->new(-file=>">".$tempfile);
     while(my $seq=$seqfile->next_seq){
+      my $qual = $seq->qual || "I" x $seq->length;
       my $seqWithQual = Bio::Seq::Quality->new(
         # TODO preserve qual values if they exist, but
         # for now, it doesn't really matter.
-        -qual=> "I" x $seq->length,
+        -qual=> $qual,
         -seq => $seq->seq,
         -id  => $seq->id,
         -verbose => -1,
@@ -183,6 +186,9 @@ sub new{
     sample     =>$$settings{sample},
     verbose    =>$$settings{verbose},
 
+    # Jellyfish-specific
+    jellyfish  =>scalar(which("jellyfish")),
+
     # Values that will be filled in after analysis
     _kmers     =>{},
   };
@@ -259,7 +265,7 @@ sub query{
     my $kmers=$self->kmers();
     $count=$$kmers{uc($query)} || 0;
   } elsif($self->{kmercounter} eq "jellyfish"){
-    open(my $queryFh, "jellyfish query ".$self->{jellyfishdb}." |") or die "ERROR: could not run jellyfish query: $!";
+    open(my $queryFh, "$$self{jellyfish} query ".$self->{jellyfishdb}." |") or die "ERROR: could not run jellyfish query: $!";
     my $db=$self->{jellyfishdb};
     my $tmp=`jellyfish query $db $query`;
     die "ERROR: could not run jellyfish query" if $?;
@@ -303,8 +309,8 @@ sub histogramJellyfish{
 
   # Run jellyfish histo
   my $jellyfishXopts = join(" ","-t", $self->{numcpus}, "-o", $self->{histfile}, $self->{jellyfishdb});
-  system("jellyfish histo $jellyfishXopts");
-  die "ERROR with jellyfish histo" if $?;
+  system("$$self{jellyfish} histo $jellyfishXopts");
+  croak "ERROR with jellyfish histo" if $?;
   
   # Read the output file
   my @hist=(0);
@@ -537,7 +543,7 @@ sub subtract{
   my($self,$other)=@_;
 
   if(!$self->_checkCompatibility($other)){
-    die;
+    croak "ERROR: trying to subtract two incompatible ".ref($self)." objects.";
   }
 
   my %subtractKmers = %{ $self->kmers };
@@ -556,7 +562,7 @@ sub _checkCompatibility{
   my($self,$other)=@_;
 
   if($self->{kmerlength} != $other->{kmerlength}){
-    warn "WARNING: kmer lengths do not match\n" if($self->{verbose});
+    carp "WARNING: kmer lengths do not match\n" if($self->{verbose});
     return 0;
   }
 
@@ -594,7 +600,7 @@ sub countKmersJellyfish{
   if($jfVersion =~ /(jellyfish\s+)?(\d+)?/){
     my $majorVersion=$2;
     if($majorVersion < 2){
-      die "ERROR: Jellyfish v2 or greater is required";
+      die "ERROR: Jellyfish v2 or greater is required for ".ref($self);
     }
   }
   
@@ -603,14 +609,18 @@ sub countKmersJellyfish{
   # Counting
   my $jellyfishCountOptions="-s 10000000 -m $kmerlength -o $outfile -t $self->{numcpus}";
   my $uncompressedFastq="$self->{tempdir}/$basename.fastq";
+  my $zcat = which("zcat");
   if($seqfile=~/\.gz$/i){
-    system("zcat $seqfile > $uncompressedFastq"); die if $?;
-    system("jellyfish count $jellyfishCountOptions $uncompressedFastq");
+    if(!-e $zcat){
+      croak "ERROR: could not find zcat in PATH for uncompressing $seqfile";
+    }
+    system("$zcat \Q$seqfile\E > $uncompressedFastq"); croak "ERROR uncompressing $seqfile" if $?;
+    system("$$self{jellyfish} count $jellyfishCountOptions \Q$uncompressedFastq\E");
   } else {
-    system("jellyfish count $jellyfishCountOptions $seqfile");
+    system("$$self{jellyfish} count $jellyfishCountOptions \Q$seqfile\E");
   }
   close $self->{jellyfishdbFh};
-  die "Error: problem with jellyfish" if $?;
+  croak "Error: problem with jellyfish" if $?;
 }
 
 sub _dumpKmersJellyfish{
@@ -624,26 +634,11 @@ sub _dumpKmersJellyfish{
   # already have contents
   if(-s $kmerTsv < 1){
     my $lowerCount=$self->{gt}-1;
-    system("jellyfish dump --lower-count=$lowerCount --column --tab -o $kmerTsv $jfDb");
-    die if $?;
+    system("$$self{jellyfish} dump --lower-count=$lowerCount --column --tab -o \Q$kmerTsv\E \Q$jfDb\E");
+    croak "ERROR dumping kmers from jellyfish database $jfDb" if $?;
   }
 }
 
-# http://www.perlmonks.org/?node_id=761662
-sub which{
-  my($self,$exe)=@_;
-  
-  my $tool_path="";
-  for my $path ( split /:/, $ENV{PATH} ) {
-      if ( -f "$path/$exe" && -x "$path/$exe" ) {
-          $tool_path = "$path/$exe";
-          last;
-      }
-  }
-  
-  return $tool_path;
-}
-
 # Opens a fastq file in a thread-safe way.
 # Returns a file handle.
 sub openFastq{
@@ -654,7 +649,7 @@ sub openFastq{
   my($name,$dir,$ext)=fileparse($fastq,@fastqExt);
 
   if(!grep(/$ext/,@fastqExt)){
-    die "ERROR: could not read $fastq as a fastq file";
+    croak "ERROR: could not read $fastq as a fastq file";
   }
 
   # Open the file in different ways, depending on if it
@@ -663,13 +658,14 @@ sub openFastq{
   if($ext =~/\.gz$/){
     # use binary gzip if we can... why not take advantage
     # of the compiled binary's speedup?
-    if(-e "/usr/bin/gzip"){
-      open($fh,"gzip -cd $fastq | ") or die "ERROR: could not open $fastq for reading!: $!";
+    my $gzip = which('gzip');
+    if(-e $gzip){
+      open($fh,"$gzip -cd \Q$fastq\E | ") or croak "ERROR: could not open $fastq for reading!: $!";
     }else{
-      $fh=new IO::Uncompress::Gunzip($fastq) or die "ERROR: could not read $fastq: $!";
+      $fh=new IO::Uncompress::Gunzip($fastq) or croak "ERROR: could not read $fastq using native perl module IO::Uncompress::Gunzip: $!";
     }
   } else {
-    open($fh,"<",$fastq) or die "ERROR: could not open $fastq for reading!: $!";
+    open($fh,"<",$fastq) or croak "ERROR: could not open $fastq for reading!: $!";
   }
 
   return $fh;
